{
  The simple demo of spectrum calculation
  Author:  Eugen Tarasov

  Last modified - January, 2004

  Please send all suggestions and comments to author.

  e-mail: dsplab@teworks.com
  http://www.teworks.com
}

unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Wave, ComCtrls, TeEngine, Series, TeeProcs, Chart, ExtCtrls,
  Buttons, dspFFT, Math;

type
  TfmMain = class(TForm)
    StatusBar1: TStatusBar;
    Panel1: TPanel;
    btnOpen: TBitBtn;
    PLeftChannel: TPanel;
    Bevel1: TBevel;
    Bevel2: TBevel;
    Splitter1: TSplitter;
    Panel3: TPanel;
    Panel4: TPanel;
    Panel5: TPanel;
    Panel6: TPanel;
    Memo1: TMemo;
    Chart1: TChart;
    Series1: TFastLineSeries;
    cbResolution: TComboBox;
    Label1: TLabel;
    tbPosition: TTrackBar;
    Label2: TLabel;
    FFTLeft: TdspFFT;
    OpenDialog1: TOpenDialog;
    PRightChannel: TPanel;
    Splitter2: TSplitter;
    Panel8: TPanel;
    Panel9: TPanel;
    Memo2: TMemo;
    Panel10: TPanel;
    Panel11: TPanel;
    Chart2: TChart;
    Series2: TFastLineSeries;
    FFTRight: TdspFFT;
    Label3: TLabel;
    procedure btnOpenClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure cbResolutionChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    procedure OnMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  public
    Wave: TWaveFileStream;

    procedure CalculateSpectrum;
    //procedure Overlap;
  end;

var
  fmMain: TfmMain;

implementation

{$R *.DFM}

//uses dspTypes;

type
  // Does trackbar more flexible
  TMyTrackBar = class(TTrackBar)
  public
    property OnMouseUp;
    property OnMouseMove;
  end;


procedure TfmMain.OnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CalculateSpectrum;
end;


// Improve navigation
procedure TfmMain.OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  tbPosition.Hint:= IntToStr(tbPosition.Position);
  Application.ActivateHint(Mouse.CursorPos);
end;


procedure TfmMain.CalculateSpectrum;
var
  I: Integer;
  DataSize: Integer;
  DataBuffer: Pointer;
  P16: PSmallInt;
  P8: PByte;
  Frequency, FreqResolution: Single;

  function F2Str(V: Single): String;
  begin
    Str(V:8:1, Result);
  end;

  procedure DisplaySpectrum(FFT: TdspFFT; Series: TFastLineSeries; Memo: TMemo);
  var I: Integer;
  begin
    FFT.FFT;
    FFT.CalculateMagnitudes;
    Series.Clear;
    {Series1.BeginUpdate; // TChart professional version only
    try}
      Memo.Lines.BeginUpdate;
      try
        Memo.Lines.Clear;
        Memo.Lines.Add('Position: ' + IntToStr(tbPosition.Position));
        Memo.Lines.Add('Frequency (Hz): Amplitude');

        for I:= 0 to FFT.BufferSize2 - 1 do
        begin
          Frequency:= I * FreqResolution;
          Series.AddXY(Frequency, FFT.RealOut[I]);
          Memo.Lines.Add(F2Str(Frequency) + ': ' + F2Str(FFT.RealOut[I]));
        end;
      finally
        Memo.Lines.EndUpdate;
      end;
    {finally // TChart professional version only
      Series1.EndUpdate;
    end;}
  end;

begin
  // 1. Allocate buffer for Wave Data
  // 2. Read data from wave stream
  // 3. Copy data to the FFT
  // 4. Calculate FFT & Spectrum
  // 5. Display spectrum

  if Wave = nil  then Exit; // Check if wave file is open
  Screen.Cursor:= crHourGlass; // Process may be long
  try
    DataSize:= Min(FFTLeft.BufferSize, (tbPosition.Max - tbPosition.Position) div Wave.Format.nBlockAlign);
    GetMem(DataBuffer, DataSize * Wave.Format.nBlockAlign);
    try
      Wave.Position:= (tbPosition.Position div Wave.Format.nBlockAlign) * Wave.Format.nBlockAlign; // Align data
      DataSize:= Wave.Read(DataBuffer^, DataSize * Wave.Format.nBlockAlign) div Wave.Format.nBlockAlign;

      FFTLeft.Clear; FFTRight.Clear;
      if Wave.Format.wBitsPerSample = 8 then // 8 bit
      begin
        P8:= DataBuffer;
        if Wave.Format.nChannels = 1 then // 8 bit Mono
        begin
          for I:= 0 to DataSize - 1 do
          begin
            FFTRight.RealIn[I]:= P8^; Inc(P8);
          end;
        end else // 8 bit stereo
        begin
          for I:= 0 to DataSize - 1 do
          begin
            FFTLeft.RealIn[I]:= P8^; Inc(P8);
            FFTRight.RealIn[I]:= P8^; Inc(P8);
          end;
        end;
      end else // 16 bit
      begin
        P16:= DataBuffer;
        if Wave.Format.nChannels = 1 then // 16 bit Mono
        begin
          for I:= 0 to DataSize - 1 do
          begin
            FFTRight.RealIn[I]:= P16^; Inc(P16);
          end;
        end else // 16 bit stereo
        begin
          for I:= 0 to DataSize - 1 do
          begin
            FFTLeft.RealIn[I]:= P16^; Inc(P16);
            FFTRight.RealIn[I]:= P16^; Inc(P16);
          end;
        end;
      end;
    finally
      FreeMem(DataBuffer);
    end;

    // Display spectrum
    FreqResolution:= Wave.Format.nSamplesPerSec / FFTLeft.BufferSize; // Sample rate / FFT Length
    DisplaySpectrum(FFTRight, Series2, Memo2);
    if Wave.Format.nChannels <> 1 then
      DisplaySpectrum(FFTLeft, Series1, Memo1);
  finally
    Screen.Cursor:= crDefault;
  end;
end;


// Setup some parameters
procedure TfmMain.FormCreate(Sender: TObject);
begin
  Caption:= Application.Title;
  cbResolution.ItemIndex:= 3;
  Resize;

  // Setup trackbar
  TMyTrackBar(tbPosition).OnMouseUp:= OnMouseUp;
  TMyTrackBar(tbPosition).OnMouseMove:= OnMouseMove;
end;


// Destroy Wave object
procedure TfmMain.FormDestroy(Sender: TObject);
begin
  Wave.Free;
end;


// Setup FFT length (resolution)
// Resolution = Sample rate / FFT length
procedure TfmMain.cbResolutionChange(Sender: TObject);
var
  R: Single;
  S: String;
begin
  FFTLeft.Power:= 7 + cbResolution.ItemIndex;
  FFTRight.Power:= FFTLeft.Power;

  if Wave <> nil then
  begin
    R:= Wave.Format.nSamplesPerSec / FFTLeft.BufferSize;
    Str(R:4:1, S);
    Label3.Caption:= '=> Resolution = ' + S + ' Hz';

    CalculateSpectrum;
  end else Label3.Caption:= '';
end;


// Open Wave file
procedure TfmMain.btnOpenClick(Sender: TObject);
var S: String;
begin
  if OpenDialog1.Execute then
  begin
    tbPosition.Position:= 0;
    tbPosition.Enabled:= false;
    StatusBar1.SimpleText:= ''; // clear file info

    Wave.Free;                  // recreate wave stream
    Wave:= TWaveFileStream.Create(OpenDialog1.FileName, fmOpenRead);
    Wave.Open(false);

    // check mono
    if Wave.Format.nChannels <> 1 then
    begin
      PLeftChannel.Visible:= true;
      Panel11.Caption:= 'Spectrum (Right channel)';
      Panel9.Caption:= 'Values (Right Channel)';
    end else
    begin
      PLeftChannel.Visible:= false;
      Panel11.Caption:= 'Spectrum';
      Panel9.Caption:= 'Values';
    end;

    Chart1.BottomAxis.Maximum:= Wave.Format.nSamplesPerSec / 2;
    Chart2.BottomAxis.Maximum:= Chart1.BottomAxis.Maximum;

    tbPosition.Max:= Wave.Size;
    tbPosition.Enabled:= true;

    // display file information
    if Wave.Format.nChannels <> 1 then S:= 'Stereo' else S:= 'Mono';
    StatusBar1.Panels[0].Text:= ' ' +
      IntToStr(Wave.Format.nSamplesPerSec) + ' Hz, ' +
      IntToStr(Wave.Format.wBitsPerSample) + ' bit, ' + S  + ', ' +
      IntToStr(Wave.Size) + ' bytes    ';
    StatusBar1.Panels[0].Width:= StatusBar1.Canvas.TextWidth(StatusBar1.Panels[0].Text);
    StatusBar1.Panels[1].Text:= ' ' + OpenDialog1.FileName;
    StatusBar1.Hint:= ExtractFileName(StatusBar1.Panels[1].Text);

    cbResolutionChange(nil); // setup & calculate spectrum
  end;
end;


procedure TfmMain.FormResize(Sender: TObject);
begin
  PLeftChannel.Height:= (ClientHeight - Panel1.Height - StatusBar1.Height) div 2;
end;

end.
